<?php
namespace FiloBlu\Rma\Model;

use FiloBlu\ExtInventory\Helper\Data;
use FiloBlu\Rma\Api\ChangeSizeInterface;
use FiloBlu\Rma\Helper\ChangeSizeAttributeHelper;
use FiloBlu\Rma\Logger\Logger;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Customer\Api\Data\AttributeMetadataInterface;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\InventorySalesApi\Api\Data\SalesChannelInterface;
use Magento\Sales\Model\Order\ItemRepository;
use Magento\Catalog\Model\ProductRepository;
use Magento\ConfigurableProduct\Api\LinkManagementInterface;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\Controller\Result\JsonFactory;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\CatalogInventory\Api\StockStateInterface;
use Magento\Customer\Model\Session;
use Magento\Catalog\Model\ResourceModel\Product;
use FiloBlu\Rma\Helper\RmaHelper;
use Magento\ConfigurableProduct\Model\ResourceModel\Product\Type\Configurable;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Module\Manager;
use Magento\Rma\Api\RmaAttributesManagementInterface;
use Magento\Sales\Api\OrderItemRepositoryInterface;

class ChangeSize implements ChangeSizeInterface
{

    const STOCK_MIN_QTY = 'filoblu_disablerma_section/rma_group/rma_minimum_qty';

    /**
     * @var ChangeSizeAttributeHelper
     */
    protected $changeSizeAttributeHelper;

    /**
     * @var ItemRepository
     */
    protected $itemRepository;

    /**
     * @var ProductRepository
     */
    protected $productRepository;

    /**
     * @var LinkManagementInterface
     */
    protected $linkManagment;

    /**
     * @var StoreManagerInterface
     */
    protected $storeManager;

    /**
     * @var JsonFactory
     */
    protected $jsonFactory;
    /**
     * @var CollectionFactory
     */
    protected $collectionFactory;

    /**
     * @var StockStateInterface
     */
    protected $stockItem;

    /**
     * @var Session
     */
    protected $customerSession;

    /**
     * @var Logger
     */
    protected $logger;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var \Magento\Catalog\Model\ResourceModel\Product
     */
    protected $resourceModel;

    /**
     * @var Configurable
     */
    protected $configurableResourceModel;

    /**
     * @var RmaHelper
     */
    protected $rmaHelper;

    /**
     * @var ObjectManager
     */
    private $objectManager;
    /**
     * @var Manager
     */
    private $manager;
    /**
     * @var RmaAttributesManagementInterface
     */
    private $rmaAttributesManagement;
    /**
     * @var OrderItemRepositoryInterface
     */
    private $orderItemRepository;
    /**
     * @var OrderRepositoryInterface
     */
    private $orderRepository;

    /**
     * ChangeSize constructor.
     * @param ChangeSizeAttributeHelper $changeSizeAttributeHelper
     * @param ItemRepository $itemRepository
     * @param ProductRepository $productRepository
     * @param LinkManagementInterface $linkManagement
     * @param JsonFactory $jsonFactory
     * @param CollectionFactory $collectionFactory
     * @param StockStateInterface $stockItem
     * @param Session $customerSession
     * @param Product $resourceModel
     * @param Configurable $configurableResourceModel
     * @param RmaHelper $rmaHelper
     * @param ScopeConfigInterface $scopeConfig
     * @param StoreManagerInterface $storeManager
     * @param Logger $logger
     * @param Manager $manager
     * @param RmaAttributesManagementInterface $rmaAttributesManagement
     * @param OrderItemRepositoryInterface $orderItemRepository
     * @param OrderRepositoryInterface $orderRepository
     */
    public function __construct(
        ChangeSizeAttributeHelper $changeSizeAttributeHelper,
        ItemRepository $itemRepository,
        ProductRepository $productRepository,
        LinkManagementInterface $linkManagement,
        JsonFactory $jsonFactory,
        CollectionFactory $collectionFactory,
        StockStateInterface $stockItem,
        Session $customerSession,
        Product $resourceModel,
        Configurable $configurableResourceModel,
        RmaHelper $rmaHelper,
        ScopeConfigInterface $scopeConfig,
        StoreManagerInterface $storeManager,
        Logger $logger,
        Manager $manager,
        RmaAttributesManagementInterface $rmaAttributesManagement,
        OrderItemRepositoryInterface $orderItemRepository,
        OrderRepositoryInterface $orderRepository
    )
    {
        $this->changeSizeAttributeHelper =$changeSizeAttributeHelper;
        $this->itemRepository =$itemRepository;
        $this->productRepository = $productRepository;
        $this->linkManagment = $linkManagement;
        $this->jsonFactory = $jsonFactory;
        $this->collectionFactory = $collectionFactory;
        $this->stockItem = $stockItem;
        $this->customerSession =$customerSession;
        $this->scopeConfig = $scopeConfig;
        $this->resourceModel = $resourceModel;
        $this->configurableResourceModel = $configurableResourceModel;
        $this->rmaHelper = $rmaHelper;
        $this->storeManager =$storeManager;
        $this->objectManager = ObjectManager::getInstance();
        $this->logger = $logger;
        $this->manager = $manager;
        $this->rmaAttributesManagement = $rmaAttributesManagement;
        $this->orderItemRepository = $orderItemRepository;
        $this->orderRepository = $orderRepository;
    }


    /**
     * @return mixed|string
     */
    public function getLabel()
    {
        try {
            return $this->changeSizeAttributeHelper->getChangeSizeAttributeOptionId();
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
            return ChangeSizeAttributeHelper::REASON_CHANGE_SIZE_LABEL;
        }
    }


    /**
     * @param string $productIds
     * @param int $websiteId
     * @return array|mixed
     */
    public function getChangeSizeEnabledProducts(string $productIds, int $websiteId){
        try {
            $enabledProducts = [];
            if($this->changeSizeAttributeHelper->isChangeSizeEnabled($websiteId)) {

                $productIds = explode(",", $productIds);
                foreach ($productIds as $productId) {
                    $item = $this->itemRepository->get($productId);
                    $storeId = $this->storeManager->getWebsite($websiteId)->getDefaultStore()->getId();
                    $product = $this->productRepository->get($item->getSku(), false, $storeId);
                    $parent_products = $this->configurableResourceModel->getParentIdsByChild($product->getId());
                    if ($parent_products) {
                        $enabledProducts[] = $productId;
                    }
                }
            }
            return $enabledProducts;
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
            return [];
        }

    }


    /**
     * @param int $itemId
     * @param int $qty
     * @param int $resolution
     * @param int $reason
     * @param int $websiteId
     * @return array|mixed
     */
    public function getSizes(int $itemId,int $qty,int $resolution, int $reason, int $websiteId,$isRefilo=false)
    {

        //check if reason is allowed for ChangeSize Order
        if($this->rmaHelper->checkIfAllowedMagentoChangeSizeReason($reason)) {
            try {
                $storeId = $this->storeManager->getWebsite($websiteId)->getDefaultStore()->getId();
                $size = [];
                if ($resolution == $this->changeSizeAttributeHelper->getChangeSizeAttributeOptionId()) {
                    if($qty == 0){
                        $qty = 1;
                    }
                    $item = $this->itemRepository->get($itemId);
                    $product = $this->productRepository->get($item->getSku(), false, $storeId);
                    $parent_products = $this->configurableResourceModel->getParentIdsByChild($product->getId());
                    if(!$parent_products){
                        if(!$isRefilo){
                            $size = array(["value" => -1, "text" => "No Configurable Products", "selected" => false]);
                        }else{
                            $size=[];
                        }

                        return $size;
                    }
                    $parents_sku = $this->resourceModel->getProductsSku($parent_products);
                    $childProducts = $this->linkManagment->getChildren($parents_sku[0]['sku']);

                    foreach ($childProducts as $child) {
                        if ($this->isSaleable($child, $storeId,$qty)) {
                            if($child->getId() === null){
                                $child->setId($child->getIdBySku($child->getSku()));
                            }
                            $productAvailableQty = $this->getProductQtyById($child->getId(),$storeId) - $this->getMinQty();
                            if ($productAvailableQty >= $qty && $product->getId() != $child->getId()) {
                                if(!$isRefilo){
                                    $size[] = ["value" => $child->getId(), "text" => $child->getResource()->getAttribute('size')->getFrontend()->getValue($child), "selected" => false];
                                }else{
                                    $size[] = ["value" => $child->getId(), "text" => $child->getResource()->getAttribute('size')->getFrontend()->getValue($child), "qty" => $productAvailableQty];
                                }

                            }
                        }
                    }
                    if (count($size) == 0) {
                        if(!$isRefilo){
                            $size = array(["value" => 0, "text" => "No Available Size", "selected" => true]);
                        }else{
                            $size =[];
                        }

                    }
                }
                return $size;
            } catch (\Exception $e) {
                $this->logger->critical($e->getMessage());
                return [];
            }
        }else{
            if(!$isRefilo){
                return  array(["value" => 0, "text" => "The Reason Selected is not compatible with Change Size Resolution.", "selected" => true]);
            }else{
                return [];
            }

        }

    }

    /**
     * @param $product
     * @param int $storeId
     * @param int $qty
     * @return bool
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function isSaleable($product, int $storeId,int $qty)
    {
        if($this->manager->isEnabled("FiloBlu_ExtInventory")){
            $websiteId = $this->storeManager->getStore($storeId)->getWebsiteId();
            $websiteCode = $this->storeManager->getWebsite($websiteId)->getCode();
            return $this->isSaleableMultistock($product,$websiteCode,$qty);
        }
        try{
            $websiteId = $this->storeManager->getStore($storeId)->getWebsiteId();
            $products = $this->collectionFactory->create()
                ->setStoreId($storeId)
                ->addPriceData($this->customerSession->getCustomerGroupId(),$websiteId)
                ->addAttributeToFilter('sku',['eq'=>$product->getSku()])
                ->getItems();
            return $products ? true : false;
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
            return false;
        }catch (\Throwable $t){
            $this->logger->critical($t->getMessage());
            return false;
        }
    }

    /**
     * @param $product
     * @param $websiteCode
     * @param $qty
     * @return bool
     */
    public function isSaleableMultistock($product,$websiteCode,$qty){
        /** @var Data $helper */
        $helper = $this->objectManager->get(Data::class);
        $salableQty=$helper->getSalableQty($product,$websiteCode);
        return $salableQty>=$qty;
    }


    /**
     * @param $productId
     * @param $storeId
     * @return float|int
     */
    public function getProductQtyById($productId,$storeId)
    {
        try{
            if(!interface_exists('Magento\InventorySalesApi\Api\GetProductSalableQtyInterface') || !$this->manager->isEnabled('Magento_InventorySalesApi')){
                return $this->stockItem->getStockQty($productId);
            }
            $websiteCode = $this->storeManager->getWebsite($this->storeManager->getStore($storeId)->getWebsiteId())->getCode();
            $getProductSalableQty = $this->objectManager->create(\Magento\InventorySalesApi\Api\GetProductSalableQtyInterface::class);
            $stockResolver = $this->objectManager->create(\Magento\InventorySalesApi\Api\StockResolverInterface::class);
            $stockId = $stockResolver->execute(SalesChannelInterface::TYPE_WEBSITE, $websiteCode)->getStockId();
            $product = $this->productRepository->getById($productId);
            return (int)$getProductSalableQty->execute($product->getSku(), $stockId);
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
            return 0;
        }catch (\Throwable $t){
            $this->logger->critical($t->getMessage());
            return 0;
        }

    }

    /**
     * @return mixed
     */
    public function getMinQty()
    {
        return (int) $this->scopeConfig->getValue(self::STOCK_MIN_QTY);
    }

    /**
     * @param int $itemId
     * @param int $qty
     * @return \Magento\Customer\Api\Data\AttributeMetadataInterface
     */
    public function getAllAttributesWithAlternativesMetadata(int $itemId, int $qty)
    {
        $attributes = $this->rmaAttributesManagement->getAllAttributesMetadata();
        foreach ($attributes as $attributeCode) {
            try {

                if ($attributeCode->getAttributeCode() == 'reason') {
                    $reasons = $this->rmaHelper->getAllMappedReason();
                    if(count($reasons)>0){
                        $allowed =[];
                        foreach ($reasons as $v){
                            if($v['is_visible']==1){
                                $magentoReasons= $this->rmaAttributesManagement->getAttributeMetadata('reason')->getOptions();
                                foreach ($magentoReasons as $reason){
                                    if($v['brand_reason']==$reason->getValue()){
                                        $allowed[]=['label'=>$reason->getLabel(),"value"=>$reason->getValue()];
                                    }
                                }

                            }
                        }
                        $attributeCode->setOptions($allowed);
                    }
                }
                if ($attributeCode->getAttributeCode() == 'resolution') {
                    $allowed = [];
                    if (!$this->changeSizeAttributeHelper->isChangeSizeEnabled($this->storeManager->getWebsite()->getId())
                        || ($this->isChangeSizeOrder($itemId) && !$this->changeSizeAttributeHelper->canChangesizeOrderUseChangesize($this->storeManager->getWebsite()->getId()))) {

                        foreach ($attributeCode->getOptions() as $option) {
                            if ($option->getValue() != $this->changeSizeAttributeHelper->getChangeSizeAttributeOptionId()) {
                                $allowed [] = $option;
                            }
                        }
                        $attributeCode->setOptions($allowed);
                    }else{
                        foreach ($attributeCode->getOptions() as $option) {
                            if($option->getValue()==""){
                                continue;
                            }
                            if ($option->getValue() == $this->changeSizeAttributeHelper->getChangeSizeAttributeOptionId()) {
                                $defaultReasonId= $this->rmaHelper->getDefaultChangeSizeMappedReason();
                                if($defaultReasonId>0){
                                    $sizes = $this->getSizes($itemId,$qty,$option->getValue(),$defaultReasonId,$this->storeManager->getWebsite()->getId(),true);
                                    if(count($sizes)>0){
                                        $option->setOptions($sizes);
                                        $allowed[]=$option;
                                    }
                                }
                            }else{
                                $allowed[]=$option;
                            }

                        }
                        $attributeCode->setOptions($allowed);
                    }
                }
                $attributesMetadata[] = $attributeCode;
            } catch (NoSuchEntityException $e) {
                // If no such entity, skip
            }
        }

        return $attributesMetadata;

    }

    /**
     * @param $itemId
     * @return bool
     */
    private function isChangeSizeOrder($itemId){
        try{
            $orderId = $this->orderItemRepository->get($itemId)->getOrderId();
            $order =$this->orderRepository->get($orderId);
            return $this->changeSizeAttributeHelper->isChangeSizeOrder($order);
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
        }catch (\Throwable $t){
            $this->logger->critical($t->getMessage());
        }
        return false;

    }

}


